"""
Build an object to represent a Clarisse Conductor submission.

In this case the submission may contain xrefs nested to any level and we do a
pretty good job of resolving them. However, if the render errors due to xref
the localize method is the fallback.
"""

import errno

import os
import re
import shutil
import sys
import tempfile
import traceback

import cioclarisse.clarisse_config as ccfg
import cioclarisse.dependencies as deps
import cioclarisse.utils as cu
import ix

from cioclarisse.job import Job
from ciocore import conductor_submit
from cioclarisse.utils import ConductorError
from ciocore import data as coredata

from ciocore.gpath import Path
 

FILE_ATTR_HINTS = [
    ix.api.OfAttr.VISUAL_HINT_FILENAME_SAVE,
    ix.api.OfAttr.VISUAL_HINT_FILENAME_OPEN,
    ix.api.OfAttr.VISUAL_HINT_FOLDER,
]


def _get_path_line_regex():
    """
    Generate a regex to help identify filepath attributes.

    As we scan project files to replace windows paths, we use this regex which
    will be something like: r'\s+(?:filename|filename_sys|save_as)\s+"(.*)"\s+' 
    only longer.
    """
    classes = ix.application.get_factory().get_classes()
    file_attrs = []
    for klass in classes.get_classes():
        attr_count = klass.get_attribute_count()
        for i in xrange(attr_count):
            attr = klass.get_attribute(i)
            hint = attr.get_visual_hint()
            if hint in FILE_ATTR_HINTS:
                file_attrs.append(attr.get_name())

    return r"\s+(?:" + "|".join(sorted(set(file_attrs))) + r')\s+"(.*)"\s+'


def _remove_conductor():
    """
    Remove all Conductor data from the render archive.

    This ensures the render logs are not polluted by complaints about Conductor
    nodes. This can only be done in the situation where we localize contexts,
    because in that case we get to reload the scene after submission.
    """
    objects = ix.api.OfObjectArray()
    ix.application.get_factory().get_objects("ConductorJob", objects)
    for item in list(objects):
        ix.application.get_factory().remove_item(item.get_full_name())


class Submission(object):
    """
    Submission holds all data needed for a submission.

    It has potentially many Jobs, and those Jobs each have many Tasks. A
    Submission can provide the correct args to send to Conductor, or it can be
    used to create a dry run to show the user what will happen.

    A Submission also sets a list of tokens that the user can access as <angle
    bracket> tokens in order to build strings in the UI such as commands, job
    title, and (soon to be added) metadata.
    """

    def __init__(self, obj):
        """
        Collect data from the Clarisse UI.

        Collect attribute values that are common to all jobs, then call
        _set_tokens(). After _set_tokens has been called, the Submission level
        token variables are valid and calls to evaluate expressions will
        correctly resolve where those tokens have been used.
        """
        self.node = obj

        if self.node.is_kindof("ConductorJob"):
            self.nodes = [obj]
        else:
            raise NotImplementedError

        self.project_filename = ix.application.get_current_project_filename()
 
        self.tmpdir = Path(
            os.path.join(
                ix.application.get_factory().get_vars().get("CTEMP").get_string(),
                "conductor",
            )
        )
        self.render_package_path = self._get_render_package_path()
        self.local_upload = self.node.get_attribute("local_upload").get_bool()

        self.should_delete_render_package = (
            self.node.get_attribute("clean_up_render_package").get_bool()
            and self.local_upload
        )

        self.upload_only = self.node.get_attribute("upload_only").get_bool()
        self.project = self._get_project()
        self.notifications = self._get_notifications()
        self.tokens = self._set_tokens()

        self.jobs = []
        for node in self.nodes:
            job = Job(node, self.tokens, self.render_package_path)
            self.jobs.append(job)

    def _get_project(self):
        """Get the project from the attr.

        Get its ID in case the current project is no longer in the list
        of projects at conductor, throw an error.
        """


        projects = coredata.data().get("projects")
        project_att = self.node.get_attribute("conductor_project_name")
        label = project_att.get_applied_preset_label()

        found = next((p for p in projects if p== label), None)
        if not found:
            msg = 'Cannot find project "{}" at Conductor. Please ensure the PROJECT dropdown contains a valid project.'.format(label)
            raise ConductorError(msg)

        return found

    def _get_notifications(self):
        """Get notification prefs."""
        if not self.node.get_attribute("notify").get_bool():
            return None

        emails = self.node.get_attribute("email_addresses").get_string()
        return [email.strip() for email in emails.split(",") if email.strip()]

    def _set_tokens(self):
        """Env tokens are variables to help the user build expressions.

        The user interface has fields for strings such as job title,
        task command. The user can use these tokens with <angle brackets> to build those strings. Tokens at the Submission
        level are also available in Job level fields, and likewise
        tokens at the Job level are available in Task level fields.
        """
        tokens = {}

        pdir_val = ix.application.get_factory().get_vars().get("PDIR").get_string()

        tokens["pdir"] = '"{}"'.format(Path(pdir_val).posix_path(with_drive=False))

        tokens["temp_dir"] = "{}".format(self.tmpdir.posix_path(with_drive=False))
        tokens["submitter"] = self.node.get_name()

        tokens["render_package"] = '"{}"'.format(
            self.render_package_path.posix_path(with_drive=False)
        )

        tokens["project"] = self.project 

        return tokens

    def _get_render_package_path(self):
        """
        Calc the path to the render package.

        The name is not always known until
        preview/submission time because it is based on the filename. 

        It will however always show up in the preview window.

        We replace spaces in the filename because of a bug in Clarisse
        https://www.isotropix.com/user/bugtracker/376

        Returns:
            string: path
        """

        msg= 'Cannot create a submission from this file. Has it ever been saved?'
        current_filename = ix.application.get_current_project_filename()
        if not current_filename:
            raise ConductorError(msg)

        node_name =self.node.get_attribute("title").get_string()

        node_name = "".join(x for x in node_name if x.isalpha() or x.isdigit()  )


        path = os.path.splitext(current_filename)[0]

        path = os.path.join(
            os.path.dirname(path), os.path.basename(path).replace(" ", "_")
        )
 
        return Path("{}.{}.cio.project".format(path,node_name))

    def get_args(self):
        """
        Prepare the args for submission to conductor.

        Returns:
            list: list of dicts containing submission args per job.
        """

        result = []
        submission_args = {}

        submission_args["local_upload"] = self.local_upload
        submission_args["upload_only"] = self.upload_only
        
        submission_args["project"] = self.project
        submission_args["notify"] = self.notifications

        for job in self.jobs:
            args = job.get_args(self.upload_only)
            args.update(submission_args)
            result.append(args)
        return result

    def submit(self):
        """
        Submit all jobs.

        Returns:
            list: list of response dictionaries, containing response codes
            and descriptions.
        """

        submission_args = self.get_args()
        self._before_submit()

        self.remove_missing_upload_paths(submission_args)
        
        results = []
  
        for job_args in submission_args:
            try:
                remote_job = conductor_submit.Submit(job_args)
                response, response_code = remote_job.main()
                results.append({"code": response_code, "response": response})
            except BaseException:
                results.append(
                    {
                        "code": "undefined",
                        "response": "".join(
                            traceback.format_exception(*sys.exc_info())
                        ),
                    }
                )
        for result in results:
            ix.log_info(result)
 
        self._after_submit()
        return results

    def _before_submit(self):
        """
        Prepare the project files that will be shipped.

        We first write out the current project file. 
        
        Then (on Windows) we find additional referenced project 
        files and adjust paths in all of them so they may be 
        rendered on linux render nodes.
        """
        self.write_render_package()


    def write_render_package(self):
        """
        Write a package suitable for rendering.

        A render package is a project file with a special name.
        """

        self._before_write_package()

        context = ix.get_item("project:/")
        ix.log_info("Writing render package: context is '{}'".format(context))
            
        package_file = self.render_package_path.posix_path()

        with cu.disabled_app():
            success = ix.export_context_as_project_with_dependencies(context, package_file)
        
        if success:
            ix.log_info("Wrote render package '{}'".format(package_file))
        else:
            msg = "Failed to write render package file '{}'".format(package_file)
            raise ConductorError(msg)

        self._after_write_package()

        if cu.is_windows():
            ix.log_info("Windows path adjustments")
            self._linuxify_render_package()
            ix.log_info("Linuxified render project file")
        else:
            ix.log_info("Not using Windows")

        return package_file

    def _linuxify_render_package(self):
        """
        Adjust reference pasths for windows.
        """

        temp_path = os.path.join(
            tempfile.gettempdir(), next(tempfile._get_candidate_names())
        )
        shutil.copy2(self.render_package_path.posix_path(), temp_path)

        os.remove(self.render_package_path.posix_path())
        self._linuxify_file(
            temp_path, self.render_package_path.posix_path()
        )

    def _linuxify_file(self, filename, dest_path=None):
        """
        Fix paths for one file.

        If the file already has the .ct.project extension, replace it too.
        """
        path_regex = _get_path_line_regex()

        out_filename = dest_path

        with open(out_filename, "w+") as outfile:
            with open(filename, "r+") as infile:
                for line in infile:
                    outfile.write(self._replace_path(line, path_regex))

    def _replace_path(self, line, path_regex):
        """
        Detect paths in the line of text and make a replacement.

        Args:
            line (string): line from the file.

        Returns:
            string: The line, possibly with replaced path
        """

        match = re.match(path_regex, line)
        if match:
            path = Path(match.group(1), no_expand=True).posix_path(with_drive=False)

            return line.replace(match.group(1), path)

        return line

    def remove_missing_upload_paths(self, submission_args):
        """
        Alert the user of missing files. If the user doesn't want to continue
        with missing files, the result will be False. Otherwise it will be True
        and the potentially adjusted args are returned.

        Args:
            submission_args (list): list of job args.

        Returns:
           tuple (bool, adjusted args):
        """
        missing_files = []

        for job_args in submission_args:
            existing_files = []
            for path in job_args["upload_paths"]:
                if os.path.exists(path):
                    existing_files.append(path)
                else:
                    missing_files.append(path)

            job_args["upload_paths"] = existing_files

        missing_files = sorted(list( missing_files))
        if missing_files:
            ix.log_warning("Skipping missing files:")
            for f in missing_files:
                ix.log_warning(f)


    def _before_write_package(self):
        """
        Prepare to write render package.
        """
        self._prepare_temp_directory()
        self._copy_system_dependencies_to_temp()

    def _prepare_temp_directory(self):
        """
        Make sure the temp directory has a conductor subdirectory.
        """
        tmpdir = self.tmpdir.posix_path()
        try:
            os.makedirs(tmpdir)
        except OSError as ex:
            if not (ex.errno == errno.EEXIST and os.path.isdir(tmpdir)):
                raise
        ix.log_info("Prepared tmpdir '{}'".format(tmpdir))

    def _copy_system_dependencies_to_temp(self):
        """
        Copy over all system dependencies to a tmp folder.

        Wrapper scripts, config files etc. The clarisse.cfg file is special. See
        ../clarisse_config.py
        """
        for entry in deps.system_dependencies():
            if os.path.isfile(entry["src"]):
                if entry["src"].endswith(".cfg"):
                    safe_config = ccfg.legalize(entry["src"])
                    with open(entry["dest"], "w") as dest:
                        dest.write(safe_config)
                    ix.log_info(
                        "Copy with mods {} to {}".format(entry["src"], entry["dest"])
                    )
                else:
                    ix.log_info("Copy {} to {}".format(entry["src"], entry["dest"]))
                    shutil.copy(entry["src"], entry["dest"])


    def _after_submit(self):
        """Clean up, and potentially other post submission actions."""
        self._delete_render_package()

    def _after_write_package(self):
        """
        Runs operations after saving the render package.

        If we did something destructive, like localize contexts, then
        a backup will have been saved and we now reload it. This strategy
        is used because Clarisse's undo is broken when it comes to
        undoing context localization.
        """
        pass
 
    def _delete_render_package(self):
        """
        Delete the render package from disk if the user wants to.
        """
        if self.should_delete_render_package:
            render_package_file = self.render_package_path.posix_path()
            if os.path.exists(render_package_file):
                os.remove(render_package_file)
 

    @property
    def node_name(self):
        """node_name."""
        return self.node.get_name()

    @property
    def filename(self):
        """filename."""
        return ix.application.get_current_project_filename()

    def has_notifications(self):
        """has_notifications."""
        return bool(self.notifications)

    @property
    def email_addresses(self):
        """email_addresses."""
        if not self.has_notifications():
            return []
        return self.notifications["email"]["addresses"]



"""
Build an object to represent a Conductor job.
"""

import json
import os

import cioclarisse.dependencies as deps
import ix
from cioclarisse import frames_ui
from cioclarisse.task import Task
from ciocore.expander import Expander
from ciocore.gpath_list import PathList
from ciocore.sequence import Sequence

from ciocore import data as coredata
from ciocore.package_environment import PackageEnvironment
from cioclarisse.utils import ConductorError

class Job(object):
    """
    Class Job holds all data for one Conductor Job in Clarisse.

    Jobs are owned by a Submission and a Job owns potentially many Tasks. Like a
    Submission, it also manages a list of tokens the user can access in the
    task template and other ConductorJob attributes.
    """

    def __init__(self, node, parent_tokens, render_package_path):
        """
        Build job object for a ConductorJob node.

        After _set_tokens has been called, the Job level token variables are
        valid and calls to evaluate string attributes will correctly resolve
        where those tokens have been used.  This is why we evaluate title,
        tasks, after the call to _set_tokens()

        Args:
            node (ConductorJob): item from which to build this job object
            parent_tokens (dict): token/value pairs in the scope of the
            submission (parent) object.
            render_package_path (string): The render project file, which must be
            added to this job's dependencies.
        """

        self.node = node
        self.tasks = []
        self.sequence = self._get_sequence()
        self.sources = self._get_sources()
        self.instance = self._get_instance()

        out = self._get_output_directory()
        self.common_output_path = out["common_path"]
        self.output_paths = out["output_paths"]

        tile_width = int(self.node.get_attribute("tiles").get_long())
        self.tiles = tile_width * tile_width

        self.tokens = self._set_tokens(parent_tokens)


        sw = self._get_software()

        self.environment = sw["env"]
        self.package_ids = sw["ids"]



        self.dependencies = deps.collect(self.node)

        try:
            self.dependencies.add(render_package_path)
        except ValueError as ex:
            msg = "{} - while resolving {}".format(str(ex), render_package_path)
            raise ConductorError(msg)

        expander = Expander(**self.tokens)
        self.title = expander.evaluate(self.node.get_attribute("title").get_string())

        # TODO: Add metadata UI
        self.metadata = None

        task_att = self.node.get_attribute("task_template")
        for chunk in self.sequence["main"].chunks():
            for tile_number in range(1, self.tiles + 1):
                tile_spec = (self.tiles, tile_number)
                self.tasks.append(
                    Task(chunk, task_att, self.sources, tile_spec, self.tokens)
                )

    def _get_sources(self):
        """
        Get the images/layers, along with associated Sequence objects.

        If we are not rendering a custom range, then the sequence for
        each image may be different.

        Returns:
            list of dict: elements contain an image along with the Sequence that
            represents the image range.
        """

        images = ix.api.OfObjectArray()
        self.node.get_attribute("images_and_layers").get_values(images)

        use_custom = self.node.get_attribute("use_custom_frames").get_bool()

        # cast to list because OfObjectArray is true even when empty.
        if not list(images):
            msg = "No render outputs added. Please add images and/or layers in the Submitter"
            raise ConductorError(msg)
        seq = self.sequence["main"]
        result = []
        for image in images:
            if not use_custom:
                seq = Sequence.create(*frames_ui.image_range(image))
            result.append({"image": image, "sequence": seq})
        return result

    def _get_extra_env_vars(self):
        """
        Collect any environment specified by the user.

        Returns:
            list of dict: name, value and merge policy of user specified env vars.
        """

        result = []
        json_entries = ix.api.CoreStringArray()
        self.node.get_attribute("extra_environment").get_values(json_entries)

        for entry in [json.loads(j) for j in json_entries]:
            result.append(
                {
                    "name": entry["key"],
                    "value": os.path.expandvars(entry["value"]),
                    "merge_policy": ["append", "exclusive"][int(entry["excl"])],
                }
            )

        return result


    def _get_internal_extra_vars(self):
        return [
            {
                "name": "PYTHONHOME",
                "value": "/opt/silhouettefx/silhouette/7/silhouette-7.5.2",
                "merge_policy": "exclusive",
            },
            {"name": "CONDUCTOR_PATHHELPER", "value": 0, "merge_policy": "exclusive"},
            {
                "name": "LD_LIBRARY_PATH",
                "value": "/usr/lib/python2.7/config-x86_64-linux-gnu",
                "merge_policy": "append",
            },
        ]
 


    def _get_software(self):
        
        result = {
            "ids": [],
            "env": PackageEnvironment()
        }

        clarisse_version_att = self.node.get_attribute("clarisse_version")
        name = clarisse_version_att.get_applied_preset_label()

        tree_data = coredata.data().get("software")
        host_package = tree_data.find_by_path(name)

        result["ids"].append(host_package["package_id"])
        result["env"].extend(host_package)
        result["env"].extend(self._get_internal_extra_vars())
        result["env"].extend(self._get_extra_env_vars())

        return result

    def _get_output_directory(self):
        """
        Get the common path for all image output paths.

        NOTE: We don't really need the subpaths any longer because directory
        creation is handled in the prerender script. Don't want to mess with
        things right now though.

        Returns:
            dict: common path and list of paths below it
        """
        out_paths = PathList()

        images = ix.api.OfObjectArray()
        self.node.get_attribute("images_and_layers").get_values(images)

        for image in images:
            directory = os.path.dirname(image.get_attribute("save_as").get_string())
            try:
                out_paths.add(directory)
            except ValueError as ex:
                msg = "{} - while resolving {}".format(str(ex), directory)
                raise ConductorError(msg)
        return {"common_path": out_paths.common_path(), "output_paths": out_paths}

    def _get_instance(self):
        """
        Get everything related to the instance.

        Get the machine type, preemptible flag, and number of retries if
        preemptible. We use the key from the instance_type menu and look
        up the machine spec in the shared data where the full list of
        instance_types is stored. When exhaustion API is in effect, the
        list of available types may be dynamic, so wetell the user to
        refresh.

        Returns:
            dict: Fields to specify the render node behaviour.
        """
        instance_types = coredata.data().get("instance_types")

        label = self.node.get_attribute("instance_type").get_applied_preset_label()

        result = {
            "preemptible": self.node.get_attribute("preemptible").get_bool(),
            "retries": self.node.get_attribute("retries").get_long(),
        }

        try:
            found = next(it for it in instance_types if str(it["description"]) == label)
        except StopIteration:
            msg = 'Invalid instance type "{}". Try a refresh (Press the Connect button).'.format(label)
            raise ConductorError(msg)

        result.update(found)
        return result

    def _get_sequence(self):

        """
        Get sequence objects from the frames section of the UI.

        Returns:
            dict: main sequence and scout sequence.
        """
        return {
            "main": frames_ui.main_frame_sequence(self.node),
            "scout": frames_ui.resolved_scout_sequence(self.node),
        }

    def _set_tokens(self, parent_tokens):
        """
        Constructs angle bracket tokens dictionary.

        Collect token values for this Job and merge with those from the
        parent Submission.

        Args:
            parent_tokens (dict): tokens that were resolved in the parent
            (submission) object.

        Returns:
            dict: combined dictionary of tokens in scope for this job.
        """
        tokens = {}
        main_seq = self.sequence["main"]
        scout_seq = self.sequence["scout"]

        tokens["scout"] = str(scout_seq)
        tokens["chunksize"] = str(main_seq.chunk_size)
        tokens["chunkcount"] = str(main_seq.chunk_count())
        tokens["scoutcount"] = str(len(scout_seq or []))
        tokens["sequencelength"] = str(len(main_seq))
        tokens["sequence"] = str(main_seq)
        tokens["sequencemin"] = str(main_seq.start)
        tokens["sequencemax"] = str(main_seq.end)
        tokens["instance_type"] = self.instance["name"]
        tokens["instance"] = self.instance["description"]
        tokens["tiles"] = str(self.tiles)
        pidx = int(self.instance["preemptible"])
        tokens["preemptible"] = "preemptible" if pidx else "non-preemptible"
        tokens["retries"] = str(self.instance["retries"])
        tokens["job"] = self.node_name

        # Space delimited list of output paths are needed for a mkdir cmd. As
        # mentioned above, no longer needed now that directories are created in
        # the prerender script. Just don't want to remove right now as no time
        # to test.
        tokens["directories"] = " ".join(
            '"{}"'.format(p.posix_path(with_drive=False)) for p in self.output_paths
        )

        tokens.update(parent_tokens)
        return tokens

    def get_args(self, upload_only):
        """
        Prepare the args for submission to conductor.

        Args:
            upload_only (bool): Don't construct tasks in an upload only job.

        Returns:
            dict: This dict represents the args that are specific to this job.
            It will be joined with the submission level args like notifications,
            and project, before submitting to Conductor.
        """
        result = {}
        result["upload_paths"] = sorted([d.posix_path() for d in self.dependencies])
        result["autoretry_policy"] = (
            {"preempted": {"max_retries": self.instance["retries"]}}
            if self.instance["preemptible"]
            else {}
        )
        result["software_package_ids"] = self.package_ids 

        result["preemptible"] = self.instance["preemptible"]
        result["environment"] = dict(self.environment)
        result["scout_frames"] = ", ".join(
            [str(s) for s in self.sequence["scout"] or []]
        )
        result["output_path"] = self.common_output_path.posix_path()
        result["chunk_size"] = self.sequence["main"].chunk_size
        result["instance_type"] = self.instance["name"]
        if not upload_only:
            result["tasks_data"] = [task.data() for task in self.tasks]

        result["job_title"] = self.title
        if self.metadata:
            result["metadata"] = self.metadata
        result["priority"] = 5
        return result

    @property
    def node_name(self):
        return self.node.get_name()



"""
Handles changes to log level from an attribute on any ConductorJob item.
"""
import ix
from ciocore import loggeria

LEVEL_LIST = [loggeria.LEVEL_MAP[key] for key in loggeria.LEVELS]


def handle_log_level(_, attr):
    """
    When a node changes log level, change all nodes log levels.

    Args:
        attr (OfAttr): Attribute that changed
    """
    nodes = ix.api.OfObjectArray()
    ix.application.get_factory().get_all_objects("ConductorJob", nodes)
    level_index = attr.get_long()
    level = loggeria.LEVELS[level_index]
    logger = loggeria.get_conductor_logger()
    logger.setLevel(level)
    for obj in nodes:
        obj.get_attribute("conductor_log_level").set_long(level_index)


def refresh_log_level(nodes):
    """
    On refresh resolve log level, update other nodes to reflect the same.

    Args:
        nodes (list): This is all ConductorJob nodes and we find the log level
        of the first one that is set.
    """
    logger = loggeria.get_conductor_logger()
    attrs = [obj.get_attribute("conductor_log_level") for obj in nodes]
    try:
        attr = next(
            attr for attr in attrs if attr.get_long() != loggeria.LEVELS.index("NOTSET")
        )
        handle_log_level(None, attr)
    except StopIteration:
        level = logger.getEffectiveLevel()
        level_index = LEVEL_LIST.index(level)
        for obj in nodes:
            obj.get_attribute("conductor_log_level").set_long(level_index)



"""
Provides a window in which to add extra env vars.

This is required because Clarisse's attribute editor doesn't allow
custom UI to be embedded.
"""

import json
import ix
from cioclarisse import refresh

BTN_HEIGHT = 22
BTN_WIDTH = 100
WINDOW_LEFT = 600
WINDOW_TOP = 200
HEIGHT = 300
WIDTH = 900
PADDING = 5

SYMBOL_BUT_WIDTH = 30
CHECKBOX_WIDTH = 50

BOTTOM_BUT_WIDTH = WIDTH / 3
VAR_WDG_WIDTH = 300
VAL_WDG_WIDTH = WIDTH - (
    SYMBOL_BUT_WIDTH + CHECKBOX_WIDTH + VAR_WDG_WIDTH + (PADDING * 5)
)


class VarWidget(ix.api.GuiWidget):
    """
    This is a line item containing UI for one env var.

    An env var has a merge policy (exclusive or append), so we have a checkbox
    for that. And we have a button to delete the line. We inherit because
    GuiWidget::get_child_items is broken and we have to have a way to store the
    children.
    """

    def __init__(self, parent, y_val, entry):
        super(VarWidget, self).__init__(parent, PADDING, y_val, WIDTH, BTN_HEIGHT)
        x_val = PADDING
        self.var_widget = ix.api.GuiLineEdit(
            self, x_val, y_val, VAR_WDG_WIDTH, BTN_HEIGHT
        )
        x_val += VAR_WDG_WIDTH + PADDING
        self.val_widget = ix.api.GuiLineEdit(
            self, x_val, y_val, VAL_WDG_WIDTH, BTN_HEIGHT
        )
        x_val += VAL_WDG_WIDTH + PADDING
        self.excl_cbx = ix.api.GuiCheckbox(self, x_val, y_val, "excl")
        x_val += CHECKBOX_WIDTH + PADDING
        self.delete_btn = ix.api.GuiPushButton(
            self, x_val, y_val, SYMBOL_BUT_WIDTH, BTN_HEIGHT, "X"
        )

        self.var_widget.set_text(str(entry.get("key", "VARNAME")))
        self.val_widget.set_text(str(entry.get("value", "/some/value")))
        self.excl_cbx.set_value(bool(entry.get("excl", False)))
        self.show()

    def index(self):
        """
        Calculates the index of this widget in the list.

        Returns:
            int: The index, derived from it's y pos so that we don't have to
            maintain another variable each time a line is added or deleted.
        """
        return (self.get_y() - PADDING - BTN_HEIGHT) / BTN_HEIGHT

    def set_y_index(self, index):
        """
        Adjust this widget's height based on its list position.

         Notice here we can't simply set the height of the container.
        Widgets locations are relative to the window, not their direct
        parent. In fact the only reason to set the parent's position is
        OCD.

        Args:
            index (int): index in the list of widgets.
        """
        y_val = ((index + 1) * BTN_HEIGHT) + PADDING
        self.set_position(self.get_x(), y_val)
        self.var_widget.set_position(self.var_widget.get_x(), y_val)
        self.val_widget.set_position(self.val_widget.get_x(), y_val)
        self.delete_btn.set_position(self.delete_btn.get_x(), y_val)

    def to_json(self):
        """
        Serialize for storage in the node's attribute.

        Returns:
            string: The entry as json
        """
        return json.dumps(
            {
                "key": self.var_widget.get_text(),
                "value": self.val_widget.get_text(),
                "excl": self.excl_cbx.get_value(),
            },
            sort_keys=True,
        )


class EnvListWidget(ix.api.GuiPanel):
    """
    The panel within the window that provides scrollbars.
    """

    def __init__(self, parent, y_val):
        """
        Configure the panel and initialize an item_list.

        Since the GuiPanel::get_child_items is broken, we have to
        maintain our own list of children, and in fact this is the only
        reason for inheriting rather than using GuiPanel directly.

        Args:
            parent (GuiWidget): Widget that this widget sits in.
            y_val (int): location of top of widget.
        """
        super(EnvListWidget, self).__init__(parent, 0, y_val, WIDTH, HEIGHT)
        self.set_scroll_enabled(True)
        self.set_bottom_toolbar_visible(False)
        self.set_top_toolbar_visible(False)
        self.item_list = []

    def _items_height(self):
        """
        Calculate the total height of items.

        Need this in order to add new items at the bottom, and to tell
        the scollable area how big the virtual space is.

        Returns:
            int: height of items.
        """
        return PADDING + ((len(self.item_list) + 1) * BTN_HEIGHT)

    def add_entries(self, *entries):
        """
        Add a line for each entry and hook up the delete button event.
        """
        for entry in entries:
            y_val = self._items_height()
            wdg = VarWidget(self, y_val, entry)
            self.item_list.append(wdg)
            self.connect(wdg.delete_btn, "EVT_ID_PUSH_BUTTON_CLICK", self.on_remove_but)

        self.set_virtual_size(self.get_width(), self._items_height())

    def on_remove_but(self, sender, _):
        """
        Remove from view.

        We have to keep the item_list in sync.

        Args:
            sender (GuiButton): button that sent the event.
        """
        parent = sender.get_parent()
        i = parent.index()
        parent.hide()
        parent.destroy()
        self.item_list.pop(i)
        self._arrange_items()

    def _arrange_items(self):
        """
        Stack the items up correctly
        """
        for i, item in enumerate(self.item_list):
            item.set_y_index(i)


class EnvWindow(ix.api.GuiWindow):
    """
    The entire window.

    Holds the panel plus buttons to Add, Go, Cancel and so on.
    """

    def __init__(self, node):
        self.node = node
        window_height = HEIGHT + (BTN_HEIGHT * 2)

        super(EnvWindow, self).__init__(
            ix.application.get_event_window(),
            WINDOW_LEFT,
            WINDOW_TOP,
            WIDTH,
            window_height,
            "Extra environment",
        )

        current_y = 0
        self.add_but = ix.api.GuiPushButton(
            self, (WIDTH - BTN_WIDTH), current_y, BTN_WIDTH, BTN_HEIGHT, "Add"
        )
        self.connect(self.add_but, "EVT_ID_PUSH_BUTTON_CLICK", self.on_add_but)

        current_y += BTN_HEIGHT
        self.panel = EnvListWidget(self, BTN_HEIGHT)
        current_y += HEIGHT

        self.close_but = ix.api.GuiPushButton(
            self, 0, current_y, BOTTOM_BUT_WIDTH, BTN_HEIGHT, "Close"
        )
        self.connect(self.close_but, "EVT_ID_PUSH_BUTTON_CLICK", self.on_close_but)

        self.apply_but = ix.api.GuiPushButton(
            self, BOTTOM_BUT_WIDTH, current_y, BOTTOM_BUT_WIDTH, BTN_HEIGHT, "Apply"
        )
        self.connect(self.apply_but, "EVT_ID_PUSH_BUTTON_CLICK", self.on_apply_but)

        self.go_but = ix.api.GuiPushButton(
            self,
            (WIDTH - BOTTOM_BUT_WIDTH),
            current_y,
            BOTTOM_BUT_WIDTH,
            BTN_HEIGHT,
            "Go",
        )
        self.connect(self.go_but, "EVT_ID_PUSH_BUTTON_CLICK", self.on_go_but)

        self.set_resizable(False)

    def on_add_but(self, sender, eventid):
        """
        Responds to add button press by adding an empty entry.
        """
        self.panel.add_entries({})

    def on_close_but(self, sender, eventid):
        """
        Hides only.
        """
        self.hide()

    def on_apply_but(self, sender, eventid):
        """
        Saves values on the attribute and keep the window visible.
        """
        self._apply()

    def on_go_but(self, sender, eventid):
        """
        Saves values on the attribute and hide(destroy) the window.
        """
        self._apply()
        self.hide()

    def _apply(self):
        """
        The real work of saving values on the attribute.
        """
        attr = self.node.get_attribute("extra_environment")
        attr.remove_all()
        for item in self.panel.item_list:
            attr.add_string(item.to_json())
        refresh.force_ae_refresh(self.node)


def build(*args):
    """
    Show the window.

    Populate it with existing entries from the extra_env attribute. The
    window is shown in modal mode so we don't keep losing the damn thing
    behind other stuff.

    Listen for events until the window is hidden.
    """
    node = args[0]
    win = EnvWindow(node)

    attr = node.get_attribute("extra_environment")
    json_entries = ix.api.CoreStringArray()
    attr.get_values(json_entries)
    entries = [json.loads(entry) for entry in json_entries]
    if not entries:
        entries = [{}]
    win.panel.add_entries(*entries)

    win.show_modal()
    while win.is_shown():
        ix.application.check_for_events()



"""
Provide UI to specify what events to notify users of.

Currently only email notifications.
"""

import re

import ix

# Simple loose email regex, matches 1 email address.
SIMPLE_EMAIL_RE = re.compile(r"^\S+@\S+$")


def handle_email_addresses(obj, _):
    """
    Validate email addresses when attribute changes.

    Args:
        obj (ConductorJob): Item from which to get notification data
    """
    val = obj.get_attribute("email_addresses").get_string().strip(",").strip()
    result = bool(val)
    for address in [x.strip() for x in val.split(",")]:
        if not SIMPLE_EMAIL_RE.match(address):
            result = False
            break
    if not result:
        ix.log_warning("Email addresses are invalid.")


def notify_changed(obj, attr):
    """
    Dim the email field based on toggle value.

    Args:
        obj (ConductorJob): Item in question
        attr (OfAttr): The toggle attribute
    """
    obj.get_attribute("email_addresses").set_read_only(not attr.get_bool())



"""Script to be added to Clarisse's list of startup scripts.

It registers the conductor_job class in Clarisse.
"""

import ix
import sys
import os

msg = """
The variable CIO_DIR is not defined in your config.env file.
It should point to the Conductor install directory, e.g. ~/Conductor/.
To fix this, please run the post_install script: `python <conductor_location>/cioclarisse/post_install.py`
"""

CIO_DIR = os.environ.get("CIO_DIR")
if not CIO_DIR:
    ix.log_error(msg)
    sys.exit()

sys.path.insert(0,  CIO_DIR)

from ciocore import data as coredata
from cioclarisse.conductor_job import ConductorJob

coredata.init(product="clarisse")
ix.api.ModuleScriptedClass.register_scripted_class(
    ix.application, "ConductorJob", ConductorJob()
)



"""
Pre render script to run inside clarisse on the render node.
Ensures images are renderable for the given range.
Ensures directories exist for renders.
"""
import argparse
import errno
import os
import re
import sys
from contextlib import contextmanager

import ix

LETTER_RX = re.compile(r"^([a-zA-Z]):")


def main():
    """
    Prepare this project to be rendered.
    """
    desc = "Prepare a Clarisse project file for rendering on Conductor"
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument("-images", nargs="+", type=str, help="Image object names.")
    options, _ = parser.parse_known_args()
    ensure_image_directories(options.images)

def ensure_image_directories(images):
    """
    Create directories in preparation for image output.

    Clarisse fails to render if the destination
    directories don't exist.

    Args:
        images (list): Clarisse paths to all images
    """

    ix.log_info("Ensure directories exist for images")
    directories = []
    for image_path in images:
        image = ix.get_item(image_path)
        directory = os.path.dirname(image.get_attribute("save_as").get_string())
        directories.append(directory)
        sys.stdout.write("{} save to disk at: {}\n".format(image_path, directory))
    mkdir_p(list(set(directories)))


def mkdir_p(dirs):
    """
    Make directories unless they already exist.

    Args:
        dirs (list): directories to create
    """
    for d in dirs:
        sys.stdout.write("Ensure directory: {} exists\n".format(d))
        try:
            os.makedirs(d)
            sys.stdout.write("Successfully made Directory:{}\n".format(d))
        except OSError as ex:
            if ex.errno == errno.EEXIST and os.path.isdir(d):
                sys.stdout.write("Directory exists: {}\n".format(d))
                pass
            else:
                sys.stderr.write(
                    "There's something wrong with the path: ('{}'), : {}\n".format(
                        d, os.strerror(ex.errno)
                    )
                )
                raise


main()



"""
Responds to events in the machines section of the ConductorJob attribute editor.

NOTE: In a future version we will also save the name of the instance type. This
    is because the value is numeric, but the menu is built dynamically from the
    list of instance types. So by stashing the name we can try to repair it if
    the list changes order between sessions.

"""
from ciocore import data as coredata
from cioclarisse import const as k

def update(obj):
    """
    Rebuilds the instance types menu. 

    Args:
        obj (ConductorJob): Item on which to rebuild menu.
    """

    instance_type_att = obj.get_attribute("instance_type")
    instance_type_att.remove_all_presets()

    if not coredata.valid():
        instance_type_att.add_preset(k.NOT_CONNECTED,  "0")
        return

    instance_types = coredata.data().get("instance_types")
    for i, instance_type in enumerate(instance_types):
        instance_type_att.add_preset(
            instance_type["description"].encode("utf-8"),  str(i))



"""Provide a window in which to display a preview submission.

This is required because Clarisse's attribute editor doesn't allow
custom UI to be embedded.
"""

import json
import os

import cioclarisse.utils as cu
import ix

C_LEFT = ix.api.GuiWidget.CONSTRAINT_LEFT
C_TOP = ix.api.GuiWidget.CONSTRAINT_TOP
C_RIGHT = ix.api.GuiWidget.CONSTRAINT_RIGHT
C_BOTTOM = ix.api.GuiWidget.CONSTRAINT_BOTTOM

BTN_HEIGHT = 22
BTN_WIDTH = 100
WINDOW_LEFT = 600
WINDOW_TOP = 200
HEIGHT = 500
WIDTH = 800
PADDING = 5

SYMBOL_BUT_WIDTH = 30
CHECKBOX_WIDTH = 50

BOTTOM_BUT_WIDTH = WIDTH / 3


def show_submission_responses(responses):
    """
    Display submission responese in a window.

    Args:
        responses (list of dict): elements contain response codes and descriptsions
    """

    success_jobs = [
        response["response"]["uri"]
        for response in responses
        if response.get("code") == 201
    ]

    messages = []
    if success_jobs:
        success_msg = ", ".join(success_jobs)
        messages.append("Successful submissions\n{}".format(success_msg))

    num_failed = len([response for response in responses if response.get("code") > 201])

    if num_failed:
        messages.append(
            "Number of failed/cancelled submissions: {:d}".format(num_failed)
        )

    if messages:
        msg = "\n".join(messages)
    else:
        msg = "No jobs were submitted"

    ix.application.message_box(
        msg,
        "Conductor Submission: Info",
        ix.api.AppDialog.yes(),
        ix.api.AppDialog.STYLE_OK,
    )


class PreviewWindow(ix.api.GuiWindow):
    """The entire window.

    Holds the panel plus buttons to Submit or Cancel.
    """

    def __init__(self, submission):
        window_height = HEIGHT + BTN_HEIGHT

        super(PreviewWindow, self).__init__(
            ix.application.get_event_window(),
            WINDOW_LEFT,
            WINDOW_TOP,
            WIDTH,
            window_height,
            "Submission Preview",
        )

        self.submission = submission
        self.text_widget = ix.api.GuiTextEdit(self, 0, 0, WIDTH, HEIGHT)
        self.text_widget.set_constraints(C_LEFT, C_TOP, C_RIGHT, C_BOTTOM)
        self.text_widget.set_read_only(True)
        self.close_but = ix.api.GuiPushButton(
            self, 0, HEIGHT, BOTTOM_BUT_WIDTH, BTN_HEIGHT, "Close"
        )
        self.close_but.set_constraints(C_LEFT, C_BOTTOM, C_LEFT, C_BOTTOM)
        self.connect(self.close_but, "EVT_ID_PUSH_BUTTON_CLICK", self.on_close_but)

        self.spacer_but = ix.api.GuiPushButton(
            self, BOTTOM_BUT_WIDTH, HEIGHT, BOTTOM_BUT_WIDTH*2, BTN_HEIGHT, ""
        )
        self.spacer_but.set_constraints(C_LEFT, C_BOTTOM, C_RIGHT, C_BOTTOM)
        self.spacer_but.set_enable(False)

        self._populate()

    def _populate(self):
        """
        Put the submission args in the window.
        """
        submission_args = self.submission.get_args()
        json_jobs = json.dumps(submission_args, indent=3, sort_keys=True)
        self.text_widget.set_text(json_jobs)

    def on_close_but(self, sender, eventid):
        """
        Hide UI so that the event loop exits and window is destroyed.
        """
        self.hide()
 


def build(submission, **kw):
    """
    Show and populate the preview window.

    Populate it with submission args for each job.

    Args:
        submission (Submission): Submission object
    """
    win = PreviewWindow(submission)

    win.show_modal()
    while win.is_shown():
        ix.application.check_for_events()



"""
Contexts for performing long or delicate functions.
"""
import os
from contextlib import contextmanager

import ix

class ConductorError(Exception):
    pass


@contextmanager
def waiting_cursor():
    """
    Perform some function with the wait cursor showing.
    """
    clarisse_win = ix.application.get_event_window()
    old_cursor = clarisse_win.get_mouse_cursor()
    clarisse_win.set_mouse_cursor(ix.api.Gui.MOUSE_CURSOR_WAIT)
    yield
    clarisse_win.set_mouse_cursor(old_cursor)


@contextmanager
def disabled_app():
    """
    Disble the app to perform some function.
    """
    app = ix.application
    app.disable()
    yield
    app.enable()


def is_windows():
    return os.name == "nt"



"""
Build an object to represent a task to be run on a render node.
"""

from ciocore.expander import Expander


class Task(object):
    """A Task has a command and the list of frames.

    It provides a handful of tokens that can be used to construct the
    command.
    """

    def __init__(self, chunk, command_attr, sources, tile_spec, parent_tokens):
        """
        Build the Task object.

        Resolve the tokens and the command. After calling set_tokens, tokens
        such as start end step and so on are valid. So when we expand the
        command any tokens that were used are correctly resolved.

        Args:
            chunk (Sequence): The frames to be rendered.
            command_attr (OfAttr): The task template attribute.
            sources (list(dict(OfObject, Sequence))):  list of images along with
            a sequence that represents the frames they will be rendered for.
            tile_spec (tuple): The number of tiles and the current tile number.
            parent_tokens (dict):Angle bracket tokens.
        """

        self.chunk = chunk
        self.tile_spec = tile_spec
        self.sources = sources
        self.tokens = self._set_tokens(parent_tokens)
        expander = Expander(**self.tokens)
        self.command = expander.evaluate(command_attr.get_string())

    def _set_tokens(self, parent_tokens):
        """Env tokens at the task level.

        Task level tokens are joined with Job and submission level
        tokens so that all tokens are available when constructing the
        task command.

        As ConductorJob may contain many images, and those
        images may specify different frame ranges. The Clarisse render
        command `cnode` handles this by specifying as many frame-specs
        as there are images. For example `cnode -images img1 img2
        -frames 1-20 30-40`. For Conductor to handle this and calculate
        a command per chunk, it must intersect the chunk with the frame
        range from each image, and apply this in the <chunks> token,
        which will be used by the task_template.
        """

        tokens = {}

        chunks = []
        image_names = []
        for source in self.sources:
            intersection = self.chunk.intersection(source["sequence"])
            if intersection:
                chunks.append(intersection.to(":", "%", "; "))
                image_names.append(source["image"].get_full_name())

        tokens["tile_number"] = str(self.tile_spec[1])
        tokens["chunks"] = " ".join(chunks)
        tokens["sources"] = " ".join(image_names)
        tokens["chunklength"] = str(len(self.chunk))
        tokens["chunkstart"] = str(self.chunk.start)
        tokens["chunkend"] = str(self.chunk.end)

        tokens.update(parent_tokens)
        return tokens

    def data(self):
        """
        The command and frame spec.

        Returns:
            dict: Frames with which to determine if its a scout frame, and the
            resolved command itself.
        """
        return {"frames": str(self.chunk), "command": self.command}



"""Handle button presses to submit and preview jobs.

Preview, open a window containing the submission script JSON, and
eventually also the structure of the submission and the JSON objects
that will be sent to Conductor.

Submit, send jobs straight to Conductor.
"""
import ix
from cioclarisse import utils
from cioclarisse import preview_ui
from cioclarisse.submission import Submission
from cioclarisse import validation
from ciocore.validator  import ValidationError

 
def submit(*args):
    """
    Validate and submit directly.
    """
    node = args[0]
    try:
        validation.run(node)
    except ValidationError as ex:
        ix.log_error(str(ex))
        return

    with utils.waiting_cursor():
        submission = Submission(node)
        responses = submission.submit()
    preview_ui.show_submission_responses(responses)

def write_render_package(*args):
    node = args[0]
    with utils.waiting_cursor():
        submission = Submission(node)
        submission.write_render_package()

def preview(*args):
    """
    Validate and show the script in a panel.

    Submission can be invoked from the preview panel.
    """
    node = args[0]
    with utils.waiting_cursor():
        submission = Submission(node)
    preview_ui.build(submission)



"""
Provide a window in which to display a Validation issues.
"""

import ix


C_LEFT = ix.api.GuiWidget.CONSTRAINT_LEFT
C_TOP = ix.api.GuiWidget.CONSTRAINT_TOP
C_RIGHT = ix.api.GuiWidget.CONSTRAINT_RIGHT
C_BOTTOM = ix.api.GuiWidget.CONSTRAINT_BOTTOM

BTN_HEIGHT = 22
BTN_WIDTH = 100
WINDOW_LEFT = 600
WINDOW_TOP = 200
HEIGHT = 500
WIDTH = 800
PADDING = 5
TITLE_PANEL_HEIGHT = 90


SYMBOL_BUT_WIDTH = 30
CHECKBOX_WIDTH = 50

BOTTOM_BUT_WIDTH = WIDTH / 3


class ValidationWindow(ix.api.GuiWindow):
    """
    The entire window.

    Holds the panel plus buttons to Continue or Cancel.
    """

    def __init__(self, warnings, notices):
        window_height = HEIGHT + BTN_HEIGHT

        super(ValidationWindow, self).__init__(
            ix.application.get_event_window(),
            WINDOW_LEFT,
            WINDOW_TOP,
            WIDTH,
            window_height,
            "Validation",
        )

        self.result = False
        self.warnings =warnings
        self.notices =notices
 
        self.title_widget = ix.api.GuiTextEdit(self, 0, 0, WIDTH, TITLE_PANEL_HEIGHT)
        self.title_widget.set_constraints(C_LEFT, C_TOP, C_RIGHT, C_TOP)
        self.title_widget.set_read_only(True)

        message = ""
        if warnings:
            message += "There are some warnings. Please study them to be sure they don't affect your render.\n\n"
        if notices:
            message += "Please check the notices. The may provide information that affects your render.\n\n"
        message += "You may continue the submission." 

        self.title_widget.set_text(message)
        self.title_widget.set_font_size(14)

        self.text_widget = ix.api.GuiTextEdit(
            self, 0, TITLE_PANEL_HEIGHT, WIDTH, HEIGHT - TITLE_PANEL_HEIGHT
        )
        self.text_widget.set_constraints(C_LEFT, C_TOP, C_RIGHT, C_BOTTOM)

        self.text_widget.set_read_only(True)
        self.close_but = ix.api.GuiPushButton(
            self, 0, HEIGHT, BOTTOM_BUT_WIDTH, BTN_HEIGHT, "Cancel"
        )
        self.close_but.set_constraints(C_LEFT, C_BOTTOM, C_LEFT, C_BOTTOM)
        self.connect(self.close_but, "EVT_ID_PUSH_BUTTON_CLICK", self.on_close_but)


        spacer_width = BOTTOM_BUT_WIDTH

        self.spacer_but = ix.api.GuiPushButton(
            self, BOTTOM_BUT_WIDTH, HEIGHT, spacer_width, BTN_HEIGHT, ""
        )
        self.spacer_but.set_constraints(C_LEFT, C_BOTTOM, C_RIGHT, C_BOTTOM)
        self.spacer_but.set_enable(False)
 
        self.go_but = ix.api.GuiPushButton(
            self,
            (WIDTH - BOTTOM_BUT_WIDTH),
            HEIGHT,
            BOTTOM_BUT_WIDTH,
            BTN_HEIGHT,
            "Continue",
        )
        self.go_but.set_constraints(C_RIGHT, C_BOTTOM, C_RIGHT, C_BOTTOM)

        self.connect(self.go_but, "EVT_ID_PUSH_BUTTON_CLICK", self.on_go_but)

        self._populate()

    def _populate(self):
        """
        Put the missing files in the window.
        """
        warnings = ["[WARN]{}".format(w) for w in self.warnings]
        notices = ["[INFO]{}".format(n) for n in self.notices]
        
        content = "\n".join(warnings + notices)
        self.text_widget.set_text(content)

    def on_close_but(self, sender, eventid):
        """
        Hide UI so that the event loop exits and window is destroyed.
        """
        self.hide()

    def on_go_but(self, sender, eventid):
        """
        Set result and hide(destroy) the window.
        """
        self.result = True

        self.hide()


def proceed( warnings, notices):
    """
    Show the window  if there are warnings or notices

    Args:
        warnings, notices (list of strings):.

    Returns:
        bool: whether or not the user wants to continue the submission or cancel.
    """
    if not (warnings or notices):
        return True

    win = ValidationWindow( warnings, notices)

    win.show_modal()

    while win.is_shown():
        ix.application.check_for_events()

    return win.result



import re

LETTER_RX = re.compile(r"([A-Z]):/")
WIN_LETTER_RX = re.compile(r"([A-Z]):\\")

BLACKLIST = [
    "recent_file_categories",
    "viewport_layout",
    "user_interface",
    "animation",
    "layout",
    "command_port",
    "scripting",
    "image_view",
    "image_history",
    "texture_view",
    "_3d_view",
    "nodal_view",
    "playblast",
    "units",
]


def legalize(filename):
    """
    Remove unnecessary stuff from a Clarisse config file.

    We have to do this to clarisse.cfg file (or whatever the user chooses to
    call it) so that it doesn't crash the rendernode. It seems that some of the
    UI configuration causes the crash, especially when shipped from  Windows.

    The algorithm relies on the config file format being pretty strict. It
    removes named blocks whose name is on its own line followed by a space and
    then an open brace. The closing brace must also be on its own line. Any
    contained blocks will also be removes, but must end the opening line with an
    open brace.

    state 0 means we are searching for blocks to remove. state >0 means we are
    moving through a block that will be removed. We count the opening and
    closing braces to know when we are done.

    TODO:
        This can be tested without loading Clarisse. Therefore, write some
        tests.
    TODO 2:
        The config file specifies a temp directory which is not writable on the
        render node. The result is permission errors in the render log and
        poslsibly a performance hit. We should search and replace it with
        something like /<output_dir>/temp.

    Args:
        filename (string): the config file

    Returns: string:
        Contents of the file, minus the blacklisted sections
    """

    state = 0
    result = []
    start_regexes = [re.compile(r"^" + x + r" {$") for x in BLACKLIST]
    open_regex = re.compile(r"^.* {$")
    close_regex = re.compile(r"^}")
    with open(filename, "r") as src_file:
        for line in src_file:
            stripped_line = line.strip()
            if state == 0:
                if any(rx.match(stripped_line) for rx in start_regexes):
                    state += 1
                    continue
                line = re.sub(LETTER_RX, "/", line)
                line = re.sub(WIN_LETTER_RX, r"\\", line)
                result.append(line)
            elif open_regex.match(stripped_line):
                state += 1
            elif close_regex.match(stripped_line):
                state -= 1

    return "".join(result)



"""
Responds to events in the frames section of the ConductorJob attribute editor.

Many functions in this module rely on the Sequence class. A Sequence is a bit
like a range() but with functions and properties more geared towards distributed
rendering of animation sequences   Please refer to the sequence.py and
sequence_test.py for more info.
"""

import ix
from ciocore.sequence import Sequence


def handle_use_custom_frames(obj, attr):
    """
    Responds to use_custom_frames changes.

    Args:
        obj (ConductorJob):
        attr (OfAttr): Attribute that changed.
    """
    hide = not attr.get_bool()
    obj.get_attribute("custom_frames").set_hidden(hide)
    update_frame_stats_message(obj)


def handle_use_scout_frames(obj, attr):
    """
    Responds to use_scout_frames changes.
    """
    obj.get_attribute("scout_frames").set_hidden(not attr.get_bool())
    update_frame_stats_message(obj)


def handle_custom_frames(obj, _):
    """
    Responds to custom_frames spec changes.
    """
    update_frame_stats_message(obj)


def handle_scout_frames(obj, _):
    """
    Responds to scout_frames spec changes.
    """
    update_frame_stats_message(obj)


def handle_chunk_size(obj, _):
    """
    Responds to chunk_size changes.
    """
    update_frame_stats_message(obj)


def handle_images(obj, _):
    """
    Responds to images or layers being added, removed, disabled.
    """
    update_frame_stats_message(obj)

def custom_frame_sequence(obj):
    """
    Generates the custom_frames sequence.

    Returns:
        Sequence: A Sequence object that represents the value in custom_frames.
        It may be a comma-separated list of progressions. Example
        1,7,10-20,30-60x3,1001.
    """
    try:

        spec = obj.get_attribute("custom_frames").get_string()
        seq = Sequence.create(
            spec, chunk_size=obj.get_attribute("chunk_size").get_long()
        )
        return seq
    except (ValueError, TypeError):
        return None


def image_range(image):
    """
    Returns the first, last, and step values from sequence attributes of an image or layer.

    Args:
        image (OfObject):  The image/layer item

    Returns:
        tuple: start, end, step
    """
    return (
        image.get_attribute("first_frame").get_long(),
        image.get_attribute("last_frame").get_long(),
        image.get_attribute("frame_step").get_long(),
    )


def union_sequence(images):
    """
    Computes the Sequence that is the union of frame ranges of all provided
    images.

    Args:
        images (list of OfObject): The images to consider.

    Returns:
        Sequence: The union Sequence.
    """
    if not images:
        return None
    rng = image_range(images[0])
    seq = Sequence.create(*rng)
    for image in images[1:]:
        rng = image_range(image)
        other = Sequence.create(*rng)
        seq = seq.union(other)
    return seq


def range_frame_sequence(obj):
    """
    Generate Sequence from value in the input images along with chunk_size attribute.

    Args:
        obj (ConductorJob): Item whose images attribute to get images from.

    Returns:
        Sequence: Sequence derived from given images.
    """

    images = ix.api.OfObjectArray()
    obj.get_attribute("images_and_layers").get_values(images)

    seq = union_sequence(list(images))
    if not seq:
        return None

    seq.chunk_size = obj.get_attribute("chunk_size").get_long()
    return seq


def main_frame_sequence(obj):
    """
    Generates Sequence that represents current settings.

    If using custom frames, then the value in the custom frames attribute are
    used, otherwise calculate the union of image sequences.

    Args:
        obj (ConductorJob): Item whose attribute to get parameters from.

    Returns:
        Sequence: Sequence that will be used for rendering.
    """

    if obj.get_attribute("use_custom_frames").get_bool():
        return custom_frame_sequence(obj)
    return range_frame_sequence(obj)


def scout_frame_sequence(obj):
    """
    Generate Sequence from value in scout_frames attribute.

    Args:
        obj (ConductorJob): Item whose attribute to get parameters from.

    Returns:
        Sequence: Sequence that represents scout frames.
    """
    try:
        spec = obj.get_attribute("scout_frames").get_string()
        return Sequence.create(spec)
    except (ValueError, TypeError):
        return None


def resolved_scout_sequence(obj):
    """
    The sub-sequence the user intends to render immediately.

    Args:
        obj (ConductorJob): Item whose attribute to get parameters from.

    Returns:
        Sequence: If do_scout is off then returning None indicates all frames
        will be rendered.

    If do_scout is on and the set of scout frames intersects the main frames,
    then return the intersection.

    If scout frames does not intersect the main frames, then the user intended
    to scout but ended up with no frames. This produces None. If it becomes a
    problem with people inadvertently rendering a whole sequence then we can
    implement a warning popup or something.

    """

    main_seq = main_frame_sequence(obj)
    if main_seq and obj.get_attribute("use_scout_frames").get_bool():
        scout_seq = scout_frame_sequence(obj)
        if scout_seq:
            return main_seq.intersection(scout_seq)
    return None


def update_frame_stats_message(obj):
    """
    Constructs an info message outlining the range that will be rendered and
    scouted.

    Args:
        obj (ConductorJob): Item whose attribute to get parameters from.
    """
    info_attr = obj.get_attribute("frames_info")

    main_seq = main_frame_sequence(obj)

    if not main_seq:
        info_attr.set_string("--")
        return

    num_frames = len(main_seq)
    scout_seq = resolved_scout_sequence(obj)
    if scout_seq:
        frame_info = "%d Scout / %d Frames" % (len(scout_seq), num_frames)
    else:
        frame_info = "%d Frames" % num_frames

    chunks = "%d Chunks" % main_seq.chunk_count()

    info_attr.set_string("{} -- {} -- {}".format(frame_info, chunks, main_seq))



"""
Collect dependencies
"""

import os
import re

import cioclarisse.utils as cu
import ix
from cioclarisse import frames_ui
from ciocore.gpath import Path
from ciocore.gpath_list import PathList
from ciocore.sequence import Sequence
from cioclarisse.utils import ConductorError

SCRIPTS_DIRECTORY = os.path.join(os.path.dirname(os.path.dirname(__file__)), "scripts")
 


# ciocnode: wrapper around cnode. Allows us to add flags and so on.

# cioprep.py: pre render script that operates in the Clarisse. Among other
# things, it removes drive letters on win.
CONDUCTOR_SCRIPTS = ["cioprep.py", "ciocnode"]

# clarisse.cfg: A copy of the users config file.
CLARISSE_CFG_FILENAME = "clarisse.cfg"

def system_dependencies():
    """
    Provides a list of system files to be sent to the render node.

    These will be copied to a directory in preparation for uploading.

    This is part of a strategy to satisfy 2 constraints.
    1. Dont store special logic on the sidecar.
    2. Don't make the render command un-runnable on the local machine.

    See docs in ciocnode and cioprep for more info.

    Returns:
        list: Each element is a source/destination pair of paths.
        [
            {"src": "/some/path.ext", "dest": "/other/path.ext"},
            ...
        ]
    """

    result = []

    conductor_tmp_dir = os.path.join(
        ix.application.get_factory().get_vars().get("CTEMP").get_string(), "conductor"
    )

    for script in CONDUCTOR_SCRIPTS:
        src_path = Path(os.path.join(SCRIPTS_DIRECTORY, script)).posix_path()
        dest_path = Path(os.path.join(conductor_tmp_dir, script)).posix_path()

        result.append({"src": src_path, "dest": dest_path})

    config_dir = (
        ix.application.get_factory()
        .get_vars()
        .get("CLARISSE_USER_CONFIG_DIR")
        .get_string()
    )

    config_src_file = Path(os.path.join(config_dir, CLARISSE_CFG_FILENAME)).posix_path()
    config_dest_file = Path(
        os.path.join(conductor_tmp_dir, CLARISSE_CFG_FILENAME)
    ).posix_path()

    result.append({"src": config_src_file, "dest": config_dest_file})

    return result


def _get_system_dependencies():
    """
    Extracts the destination side of system dependency files.

    Returns:
        PathList: list of system files to be uploaded
    """
    result = PathList()
    for entry in system_dependencies():
        try:
            result.add(entry["dest"])
        except ValueError as ex:
            msg = "{} - while resolving system_dependency: {}".format( str(ex), entry["dest"])
            raise ConductorError(msg)

    return result


def collect(obj, do_glob=True):
    """
    Collect ALL upload files in preparation for submission.

    Args:
        obj (ConductorJob): The item whose attributes define the scan.
    Returns:
        PathList: All file dependencies.
    """
 

    result = PathList()
    result.add(*_get_system_dependencies())
    result.add(*_get_extra_uploads(obj))
    result.add(*get_scan(obj))
    if do_glob:
        result.glob()
    return result


def _get_extra_uploads(obj):
    """
    Collects any files specified through the extra uploads window.

    They are stored in a list attribute on the ConductorJob item.

    Args:
        obj (ConductorJob): item being processed.

    Returns:
        PathList: Collected paths.
    """
    result = PathList()
    extras_attr = obj.get_attribute("extra_uploads")
    paths = ix.api.CoreStringArray()
    extras_attr.get_values(paths)
    for path in paths:
        try:
            result.add(path)
        except ValueError as ex:
            msg = "{} - while resolving extra upload path: {}".format(str(ex), path)
            raise ConductorError(msg)

    return result


def resolve_attr(attr):

    is_ex = attr.is_expression_enabled() and attr.is_expression_activated()
    active_value = attr.get_expression() if is_ex else  attr.get_string()
    static_value = re.sub(  r"\$(\d?)F|#+|<UDIM>", "*",  active_value)
    if not is_ex:
        return static_value
    attr.set_expression(static_value)

    # Force the expression to evaluate correctly.
    attr.activate_expression(False)
    attr.activate_expression(True)

    resolved = attr.get_string()
    # return to orig expr
    attr.set_expression(active_value)
    
    return resolved

def get_scan(obj):

    result = PathList()
    objects = ix.api.OfObjectSet()
    obj.gather_branch(objects)
    for o in objects:
        if o.is_disabled():
            continue
        for index in range(o.get_attribute_count()):
            attr =  o.get_attribute(index)
            hint = ix.api.OfAttr.get_visual_hint_name(attr.get_visual_hint())
            if hint in [ "VISUAL_HINT_FILENAME_OPEN" , "VISUAL_HINT_FOLDER" ] and not attr.get_name() == "filename_sys" :
                fn =   resolve_attr(attr) 
                if fn:
                    result.add(Path(fn)) 
                    
    return result
